"""
textDocument/completion
"""

#  Copyright (c) 2019 Seven Bridges. See LICENSE

from .lspobjects import Position
from .base import CWLLangServerBase

import logging
logger = logging.getLogger(__name__)


class Completion(CWLLangServerBase):

    def serve_textDocument_completion(self, client_query):
        params = client_query["params"]
        doc_uri = params["textDocument"]["uri"]
        position = Position(**params["position"])

        doc = self.open_documents[doc_uri]
        return doc.completion(position)



"""
textDocument/formatting
"""

#  Copyright (c) 2020 Seven Bridges. See LICENSE

from cwlformat.formatter import cwl_format

from .base import CWLLangServerBase
from .lspobjects import Position, Range, TextEdit

import logging
logger = logging.getLogger(__name__)


class Formatting(CWLLangServerBase):

    def serve_textDocument_formatting(self, client_query):
        params = client_query["params"]
        doc_uri = params["textDocument"]["uri"]
        doc = self.open_documents[doc_uri]

        if len(doc.text) == 0:
            return

        lines = doc.text.splitlines()
        _start = Position(0, 0)
        _end = Position(len(lines), len(lines[-1]))
        return [TextEdit(_range=Range(start=_start, end=_end), new_text=cwl_format(doc.text))]



#  Copyright (c) 2019 Seven Bridges. See LICENSE

from .basetype import CWLBaseType, IntelligenceContext, Intelligence, MapSubjectPredicate, TypeCheck, Match
from ..langserver.lspobjects import Range, CompletionItem, Diagnostic, DiagnosticSeverity, Hover
from ..code.intelligence import LookupNode
from ..code.yaml import yaml_to_string
from .lib import normalized_path

import logging
logger = logging.getLogger(__name__)


class CWLEnumType(CWLBaseType):

    def __init__(self, name: str, symbols: set):
        super().__init__(name)
        self.symbols = symbols
        self._hover_value = None

    def check(self, node, node_key: str=None, map_sp: MapSubjectPredicate=None) -> TypeCheck:

        if not (isinstance(node, str) or None):
            return TypeCheck(
                cwl_type=self,
                match=Match.No)
        else:
            if self.name in ["PrimitiveType", "CWLType"]:
                # Special treatment for data types
                return TypeCheck(cwl_type=CWLDataType(node, self.symbols))
            else:
                return TypeCheck(cwl_type=self)

    def parse(self,
              doc_uri: str,
              node,
              intel_context: IntelligenceContext,
              code_intel: Intelligence,
              problems: list,
              node_key: str = None,
              map_sp: MapSubjectPredicate = None,
              key_range: Range = None,
              value_range: Range = None,
              requirements=None):

        if node not in self.symbols:
            problems += [
                Diagnostic(
                    _range=value_range,
                    message=f"Expecting one of: {sorted(self.symbols)}",
                    severity=DiagnosticSeverity.Error)
            ]

        ln = LookupNode(loc=value_range)
        ln.intelligence_node = self
        code_intel.add_lookup_node(ln)

    def completion(self):
        return [CompletionItem(label=s) for s in self.symbols]


class CWLDataType(CWLEnumType):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user_types = []
        self.prefix = ""

    def parse(self,
              doc_uri: str,
              node,
              intel_context: IntelligenceContext,
              code_intel: Intelligence,
              problems: list,
              node_key: str = None,
              map_sp: MapSubjectPredicate = None,
              key_range: Range = None,
              value_range: Range = None,
              requirements=None):

        self.prefix = node
        self.user_types = code_intel.type_defs.keys()

        # de-sugar
        if node[-1] == "?":
            node = node[:-1]
        if node[-2:] == "[]":
            node = node[:-2]

        if node not in self.symbols: # Not a native type
            if node not in code_intel.type_defs: # Path doesn't match denormalized
                _norm_path = normalized_path(doc_uri, node)
                for _type in code_intel.type_defs.keys():
                    if _norm_path == normalized_path(doc_uri, _type):
                        break
                else:
                    problems += [
                        Diagnostic(
                            _range=value_range,
                            message=f"Expecting one of: {sorted(set(self.symbols).union(code_intel.type_defs.keys()))}",
                            severity=DiagnosticSeverity.Error)
                    ]

            else:
                self._hover_value = code_intel.type_defs[node]

        ln = LookupNode(loc=value_range)
        ln.intelligence_node = self
        code_intel.add_lookup_node(ln)

    def hover(self):
        if self._hover_value is not None:
            return Hover(yaml_to_string(self._hover_value), hover_type=Hover.HoverType.Code)

    def completion(self):
        return [CompletionItem(label=s) for s in self.symbols.union(self.user_types)]



"""Represents an $include or $import"""

#  Copyright (c) 2020 Seven Bridges. See LICENSE

from .basetype import CWLBaseType, IntelligenceContext, Intelligence, MapSubjectPredicate
from .linkedfiletype import CWLLinkedFile
from .linkedschemadeftype import CWLLinkedSchemaDef
from ..langserver.lspobjects import Diagnostic, DiagnosticSeverity, Range
from .lib import get_range_for_key, get_range_for_value


class CWLImportInclude(CWLBaseType):

    def __init__(self, key, import_context):
        super().__init__("Import/Include")
        self.key = key
        self.import_context = import_context

    def parse(self,
              doc_uri: str,
              node,
              intel_context: IntelligenceContext,
              code_intel: Intelligence,
              problems: list,
              node_key: str = None,
              map_sp: MapSubjectPredicate = None,
              key_range: Range = None,
              value_range: Range = None,
              requirements=None):

        if len(node) > 1:
            problems += [
                Diagnostic(
                    _range=value_range,
                    message=f"{self.key} has to be the only key",
                    severity=DiagnosticSeverity.Error)
            ]
            return

        if self.import_context == "InputRecordSchema":
            if self.key != "$import":
                problems += [
                    Diagnostic(
                        _range=value_range,
                        message=f"Expecting an $import",
                        severity=DiagnosticSeverity.Error)
                ]
                return
            inferred_type = CWLLinkedSchemaDef(prefix=node.get("$import"))
        else:
            inferred_type = CWLLinkedFile(prefix=node.get(self.key))

        inferred_type.parse(
            doc_uri=doc_uri,
            node=node.get(self.key),
            intel_context=intel_context,
            code_intel=code_intel,
            problems=problems,
            node_key=self.key,
            map_sp=map_sp,
            key_range=key_range,
            value_range=get_range_for_value(node, self.key),
            requirements=requirements)



#  Copyright (c) 2019 Seven Bridges. See LICENSE

"""Types are stored in the code_intel type dictionary. They are keyed by the URI (if off-disk) or
 the relative path if on-disk. When matching types the path is normalized. The reason why we store
 the relative path is that it makes for a better looking auto-completion."""

import pathlib
from urllib.parse import urlparse

from .linkedfiletype import CWLLinkedFile
from .basetype import IntelligenceContext, Intelligence, MapSubjectPredicate
from ..langserver.lspobjects import Diagnostic, DiagnosticSeverity, Range

import logging
logger = logging.getLogger(__name__)


class CWLLinkedSchemaDef(CWLLinkedFile):

    def parse(self,
              doc_uri: str,
              node,
              intel_context: IntelligenceContext,
              code_intel: Intelligence,
              problems: list,
              node_key: str = None,
              map_sp: MapSubjectPredicate = None,
              key_range: Range = None,
              value_range: Range = None,
              requirements=None):
        super().parse(doc_uri, node, intel_context, code_intel, problems, node_key, map_sp, key_range, value_range, requirements)

        if isinstance(self.node_dict, list):
            _type_list = self.node_dict
        elif isinstance(self.node_dict, dict):
            _type_list = [self.node_dict]
        else:
            problems += [
                Diagnostic(
                    _range=value_range,
                    message=f"Problem parsing SchemaDef file",
                    severity=DiagnosticSeverity.Error)
            ]
            return

        for _type in _type_list:
            if "name" in _type:
                name = self.prefix + "#" + _type.pop("name")
                code_intel.type_defs[name] = _type
